/*
 * Enhance Stencyl Extension [Haxe interface]
 *
 * Created by Enhance, Inc.
 * www.enhance.co
 *
 */

package;

#if (cpp && mobile && !android)
    #if (openfl < "4.0.0")
    import openfl.Lib as CPPLoader;
    #else
    import lime.system.CFFI as CPPLoader;
    #end
#end

#if android
    #if (openfl < "4.0.0")
    import openfl.utils.JNI;
    #else
    import lime.system.JNI;
    #end
#end

class Enhance
{   
    public static inline var DEFAULT_PLACEMENT:String = "default";
    public static inline var REWARDED_AD_DEFAULT_PLACEMENT:String = "neutral";

    public static inline var POSITION_TOP:String = "TOP";
    public static inline var POSITION_BOTTOM:String = "BOTTOM";

    private static inline var AD_TYPE_INTERSTITIAL:String = "INTERSTITIAL";
    private static inline var AD_TYPE_REWARDED:String = "REWARDED";
    private static inline var AD_TYPE_BANNER:String = "BANNER";
    private static inline var AD_TYPE_SPECIAL_OFFER:String = "SPECIALOFFER";
    private static inline var AD_TYPE_OFFERWALL:String = "OFFERWALL";

    private static inline var REWARD_GRANTED:Int = 0;
    private static inline var REWARD_DECLINED:Int = 1;
    private static inline var REWARD_UNAVAILABLE:Int = 2;

    private static inline var REWARD_TYPE_ITEM:String = "ITEM";
    private static inline var REWARD_TYPE_COINS:String = "COINS";

    private static inline var PERMISSION_GRANTED:Int = 0;
    private static inline var PERMISSION_REFUSED:Int = 1;

    private static inline var PURCHASE_SUCCESS:Int = 0;
    private static inline var PURCHASE_FAILED:Int = 1;

    private static inline var CONSUME_SUCCESS:Int = 0;
    private static inline var CONSUME_FAILED:Int = 1;

    private static inline var RESTORE_SUCCESS:Int = 0;
    private static inline var RESTORE_FAILED:Int = 1;

    private static inline var ON_USER_OPT_IN_REQUIRED:Int = 0;
    private static inline var ON_USER_OPT_IN_NOT_REQUIRED:Int = 1;
    private static inline var ON_DIALOG_COMPLETE:Int = 2;

    private static inline var IOS_EVENT_REWARD_GRANTED:String = "REWARD_GRANTED";
    private static inline var IOS_EVENT_REWARD_DECLINED:String = "REWARD_DECLINED";
    private static inline var IOS_EVENT_REWARD_UNAVAILABLE:String = "REWARD_UNAVAILABLE";
    private static inline var IOS_EVENT_CURRENCY_GRANTED:String = "CURRENCY_GRANTED";
    private static inline var IOS_EVENT_PERMISSION_GRANTED:String = "PERMISSION_GRANTED";
    private static inline var IOS_EVENT_PERMISSION_REFUSED:String = "PERMISSION_REFUSED";
    private static inline var IOS_EVENT_PURCHASE_SUCCESS:String = "PURCHASE_SUCCESS";
    private static inline var IOS_EVENT_PURCHASE_FAILED:String = "PURCHASE_FAILED";
    private static inline var IOS_EVENT_RESTORE_SUCCESS:String = "RESTORE_SUCCESS";
    private static inline var IOS_EVENT_RESTORE_FAILED:String = "RESTORE_FAILED";
    private static inline var IOS_EVENT_CONSUME_SUCCESS:String = "CONSUME_SUCCESS";
    private static inline var IOS_EVENT_CONSUME_FAILED:String = "CONSUME_FAILED";
    private static inline var IOS_EVENT_USER_OPT_IN_REQUIRED:String = "USER_OPT_IN_REQUIRED";
    private static inline var IOS_EVENT_USER_OPT_IN_NOT_REQUIRED:String = "USER_OPT_IN_NOT_REQUIRED";
    private static inline var IOS_EVENT_OPT_IN_DIALOG_COMPLETE:String = "OPT_IN_DIALOG_COMPLETE";

    private static inline var ACTION_PURCHASE:String = "PURCHASE";
    private static inline var ACTION_CONSUME:String = "CONSUME";
    private static inline var ACTION_SUCCESS:Int = 0;
    private static inline var ACTION_FAILED:Int = 1;

    private static inline var EXCLASS:String = "co.enhance.EnhanceExtension";
    private static inline var EXNAMESPACE:String = "enhance";

    private static var isInitialized:Bool = false;
    private static var instance:Enhance = null;

    private static var rewardCallbacks:Array<Void->Void> = [null, null, null];
    private static var permissionCallbacks:Array<Void->Void> = [null, null];
    private static var purchaseCallbacks:Array<Void->Void> = [null, null];
    private static var consumeCallbacks:Array<Void->Void> = [null, null];
    private static var restoreCallbacks:Array<Void->Void> = [null, null];
    private static var gdprCallbacks:Array<Void->Void> = [null, null, null];

    public static var lastRewardType:String = "UNKNOWN";
    public static var lastRewardValue:Int = 0;

    private static var bannerShown:Bool = false;

    private static var offerwallCallback:Int->Void = null;

    private static var hasLocalNotificationPermission:Bool = false;

    public function new() {}

    private static function init():Void
    {
        if(isInitialized) return;
        isInitialized = true;
        instance = new Enhance();

        #if android
        var androidInit:Dynamic = JNI.createStaticMethod(EXCLASS, "init", "(Lorg/haxe/lime/HaxeObject;)V", true);
        androidInit([instance]);
        #end

        #if(cpp && mobile && !android)
        var iosInit = CPPLoader.load(EXNAMESPACE, "ios_init", 1);
        iosInit(handleEventIOS);
        #end
    }

    public static function isReady(adType:String, adPlacement:String = DEFAULT_PLACEMENT):Bool
    {
        if(adType == AD_TYPE_INTERSTITIAL)
        {
            return isInterstitialReady(adPlacement);
        }

        else if(adType == AD_TYPE_REWARDED)
        {
            if(adPlacement == DEFAULT_PLACEMENT) adPlacement = REWARDED_AD_DEFAULT_PLACEMENT;
            return isRewardedAdReady(adPlacement);
        }

        else if(adType == AD_TYPE_BANNER)
        {
            return isBannerAdReady(adPlacement);
        }

        else if(adType == AD_TYPE_SPECIAL_OFFER)
        {
            return isSpecialOfferReady(adPlacement);
        }

        else if(adType == AD_TYPE_OFFERWALL)
        {
            return isOfferwallReady(adPlacement);
        }

        return false;
    }

    public static function show(adType:String, adPlacement:String = DEFAULT_PLACEMENT):Void
    {
        if(adType == AD_TYPE_INTERSTITIAL)
        {
            showInterstitialAd(adPlacement);
        }

        else if(adType == AD_TYPE_REWARDED)
        {
            if(adPlacement == DEFAULT_PLACEMENT) adPlacement = REWARDED_AD_DEFAULT_PLACEMENT;
            showRewardedAd(adPlacement);
        }

        else if(adType == AD_TYPE_BANNER)
        {
            showBannerAdWithPosition(adPlacement);
        }

        else if(adType == AD_TYPE_SPECIAL_OFFER)
        {
            showSpecialOffer(adPlacement);
        }

        else if(adType == AD_TYPE_OFFERWALL)
        {
            showOfferwall(adPlacement);
        }
    }

    public static function showWithPosition(adType:String, position:String, adPlacement:String = DEFAULT_PLACEMENT):Void
    {
        if(adType == AD_TYPE_BANNER)
        {
            showBannerAdWithPosition(adPlacement, position);
        }
    }

    public static function hide(adType:String):Void
    {
        if(adType == AD_TYPE_BANNER) 
        {
            hideBannerAd();
        }
    }

    public static function setRewardedAdCallback(responseType:Int, cb:Void->Void):Void
    {
        if(responseType >= rewardCallbacks.length) return; 
        if(cb == null) return;

        rewardCallbacks[responseType] = cb;
    }

    public static function setOfferwallCallback(responseType:Int, cb:Int->Void):Void
    {
        /* 
        it's important to call init() also here
        this function will be usually used at the very beginning of the Stencyl app
        currency callback in java/obj-c should be set asap as the event can come in at any time
        */
        if(!isInitialized) init();

        // response type doesn't matter for now as there's only one option -- currency granted
        if(cb == null) return;

        offerwallCallback = cb;
    }

    public static function setPermissionCallback(responseType:Int, cb:Void->Void):Void
    {
        if(responseType >= permissionCallbacks.length) return;
        if(cb == null) return;

        permissionCallbacks[responseType] = cb;
    }

    public static function setPurchaseCallback(responseType:Int, cb:Void->Void):Void
    {
        if(responseType >= purchaseCallbacks.length) return;
        if(cb == null) return;

        purchaseCallbacks[responseType] = cb;
    }

    // Purchase/Consume
    public static function itemAction(actionType:String, sku:String)
    {
        if(actionType == ACTION_PURCHASE)
        {
            attemptPurchase(sku);
        }

        else if(actionType == ACTION_CONSUME)
        {
            consume(sku);
        }
    }

    // Set callback for purchase/consume
    public static function setItemActionCallback(actionType:String, actionResult:Int, cb:Void->Void)
    {
        if(cb == null) return;

        if(actionType == ACTION_PURCHASE)
        {
            if(actionResult >= purchaseCallbacks.length) return;
            purchaseCallbacks[actionResult] = cb;
        }

        else if(actionType == ACTION_CONSUME)
        {
            if(actionResult >= consumeCallbacks.length) return;
            consumeCallbacks[actionResult] = cb;
        }
    }

    // Set callback for restore
    public static function setRestoreCallback(responseType:Int, cb:Void->Void):Void
    {
        if(responseType >= restoreCallbacks.length) return;
        if(cb == null) return;

        restoreCallbacks[responseType] = cb;
    }

    public static function setGdprCallback(responseType:Int, cb:Void->Void):Void
    {
        if(responseType >= gdprCallbacks.length) return;
        if(cb == null) return;

        gdprCallbacks[responseType] = cb;
    }

    // Handle events received from iOS part of the connector

    private static function handleEventIOS(eventData:Dynamic):Void
    {
        // can't handle events without Enhance class instance
        if(instance == null) return;

        #if(cpp && mobile && !android)
        var type:String = Std.string(Reflect.field(eventData, "type"));

        if(type == IOS_EVENT_REWARD_GRANTED)
        {
            var rewardValue:Int = Std.int(Reflect.field(eventData, "rewardValue"));
            var rewardType:String = Std.string(Reflect.field(eventData, "rewardType"));

            instance.onRewardGranted(rewardType, rewardValue);
        }

        else if(type == IOS_EVENT_REWARD_DECLINED)
        {
            instance.onRewardDeclined();
        }

        else if(type == IOS_EVENT_REWARD_UNAVAILABLE)
        {
            instance.onRewardUnavailable();
        }

        else if(type == IOS_EVENT_CURRENCY_GRANTED)
        {
            var currencyAmount:Int = Std.int(Reflect.field(eventData, "currencyAmount"));
            instance.onCurrencyGranted(currencyAmount);
        }

        else if(type == IOS_EVENT_PERMISSION_GRANTED)
        {
            instance.onPermissionGranted();
        }

        else if(type == IOS_EVENT_PERMISSION_REFUSED)
        {
            instance.onPermissionRefused();
        }

        else if(type == IOS_EVENT_PURCHASE_SUCCESS)
        {
            instance.onPurchaseSuccess();
        }

        else if(type == IOS_EVENT_PURCHASE_FAILED)
        {
            instance.onPurchaseFailed();
        }

        else if(type == IOS_EVENT_RESTORE_SUCCESS)
        {
            instance.onRestoreSuccess();
        }

        else if(type == IOS_EVENT_RESTORE_FAILED)
        {
            instance.onRestoreFailed();
        }

        else if(type == IOS_EVENT_CONSUME_SUCCESS)
        {
            instance.onConsumeSuccess();
        }

        else if(type == IOS_EVENT_CONSUME_FAILED)
        {
            instance.onConsumeFailed();
        }

        else if(type == IOS_EVENT_USER_OPT_IN_REQUIRED)
        {
            instance.onServiceOptInRequirement(true);
        }

        else if(type == IOS_EVENT_USER_OPT_IN_NOT_REQUIRED)
        {
            instance.onServiceOptInRequirement(false);
        }

        else if(type == IOS_EVENT_OPT_IN_DIALOG_COMPLETE)
        {
            instance.onDialogComplete();
        }
        #end
    }

    // Is Interstitial Ready //

    private static function isInterstitialReady(placement:String = DEFAULT_PLACEMENT):Bool 
    {
        if(!isInitialized) init();

        var ready:Bool = false;

        #if android
        if(androidIsInterstitialReady == null) 
        {
            androidIsInterstitialReady = JNI.createStaticMethod(EXCLASS, "isInterstitialReady", "(Ljava/lang/String;)Z", true);
        }

        ready = androidIsInterstitialReady([placement]);
        #end

        #if(cpp && mobile && !android)
        ready = iosIsInterstitialReady(placement);
        #end

        return ready;
    }

    // Show Interstitial Ad //

    private static function showInterstitialAd(placement:String = DEFAULT_PLACEMENT):Void
    {
        if(!isInitialized) init();

        #if android
        if(androidShowInterstitialAd == null) 
        {
            androidShowInterstitialAd = JNI.createStaticMethod(EXCLASS, "showInterstitialAd", "(Ljava/lang/String;)V", true);
        }

        androidShowInterstitialAd([placement]);
        #end

        #if(cpp && mobile && !android)
        iosShowInterstitialAd(placement);
        #end
    }

    // Is Rewarded Ad Ready //

    private static function isRewardedAdReady(placement:String = REWARDED_AD_DEFAULT_PLACEMENT):Bool
    {
        if(!isInitialized) init();

        var ready:Bool = false;

        #if android
        if(androidIsRewardedAdReady == null)
        {
            androidIsRewardedAdReady = JNI.createStaticMethod(EXCLASS, "isRewardedAdReady", "(Ljava/lang/String;)Z", true);
        }

        ready = androidIsRewardedAdReady([placement]);
        #end

        #if(cpp && mobile && !android)
        ready = iosIsRewardedAdReady(placement);
        #end

        return ready;
    }

    // Show Rewarded Ad //

    private static function showRewardedAd(placement:String = REWARDED_AD_DEFAULT_PLACEMENT):Void
    {
        if(!isInitialized) init();

        #if android
        if(androidShowRewardedAd == null)
        {
            androidShowRewardedAd = JNI.createStaticMethod(EXCLASS, "showRewardedAd", "(Ljava/lang/String;)V", true);
        }

        androidShowRewardedAd([placement]);
        #end

        #if(cpp && mobile && !android)
        iosShowRewardedAd(placement);
        #end
    }

    // Rewarded Ad - internal callbacks //
    // Can't be static, otherwise Java won't have an access 

    public function onRewardGranted(rewardType:String, rewardValue:Int):Void
    {
        Enhance.lastRewardType = rewardType;
        Enhance.lastRewardValue = rewardValue;

        if(rewardCallbacks[REWARD_GRANTED] != null) rewardCallbacks[REWARD_GRANTED]();
    }

    public function onRewardDeclined():Void
    {
        if(rewardCallbacks[REWARD_DECLINED] != null) rewardCallbacks[REWARD_DECLINED]();
    }

    public function onRewardUnavailable():Void
    {
        if(rewardCallbacks[REWARD_UNAVAILABLE] != null) rewardCallbacks[REWARD_UNAVAILABLE]();
    } 

    // Is Banner Ad Ready //

    private static function isBannerAdReady(placement:String = DEFAULT_PLACEMENT):Bool
    {
        if(!isInitialized) init();

        var ready:Bool = false;

        #if android
        if(androidIsBannerAdReady == null)
        {
            androidIsBannerAdReady = JNI.createStaticMethod(EXCLASS, "isBannerAdReady", "(Ljava/lang/String;)Z", true);
        }

        ready = androidIsBannerAdReady([placement]);
        #end

        #if(cpp && mobile && !android)
        ready = iosIsBannerAdReady(placement);
        #end

        return ready;
    }

    // Show Banner Ad //

    private static function showBannerAdWithPosition(placement:String = DEFAULT_PLACEMENT, position:String = POSITION_BOTTOM):Void
    {
        if(!isInitialized) init();

        if(bannerShown) return;

        bannerShown = true;

        #if android
        if(androidShowBannerAdWithPosition == null)
        {
            androidShowBannerAdWithPosition = JNI.createStaticMethod(EXCLASS, "showBannerAdWithPosition", "(Ljava/lang/String;Ljava/lang/String;)V", true);
        }

        androidShowBannerAdWithPosition([placement, position]);
        #end

        #if(cpp && mobile && !android)
        iosShowBannerAdWithPosition(placement, position);
        #end
    }

    // Hide Banner Ad //

    private static function hideBannerAd():Void
    {
        if(!isInitialized) init();

        if(!bannerShown) return;

        bannerShown = false;

        #if android
        if(androidHideBannerAd == null)
        {
            androidHideBannerAd = JNI.createStaticMethod(EXCLASS, "hideBannerAd", "()V", true);
        }

        androidHideBannerAd();
        #end

        #if(cpp && mobile && !android)
        iosHideBannerAd();
        #end
    }

    // Is Special Offer Ready //

    private static function isSpecialOfferReady(placement:String = DEFAULT_PLACEMENT):Bool
    {
        if(!isInitialized) init();

        var ready:Bool = false;

        #if android
        if(androidIsSpecialOfferReady == null)
        {
            androidIsSpecialOfferReady = JNI.createStaticMethod(EXCLASS, "isSpecialOfferReady", "(Ljava/lang/String;)Z", true);
        }

        ready = androidIsSpecialOfferReady([placement]);
        #end

        #if(cpp && mobile && !android)
        ready = iosIsSpecialOfferReady(placement);
        #end

        return ready;
    }

    // Show Special Offer //

    private static function showSpecialOffer(placement:String = DEFAULT_PLACEMENT):Void
    {
        if(!isInitialized) init();

        #if android
        if(androidShowSpecialOffer == null)
        {
            androidShowSpecialOffer = JNI.createStaticMethod(EXCLASS, "showSpecialOffer", "(Ljava/lang/String;)V", true);
        }

        androidShowSpecialOffer([placement]);
        #end

        #if(cpp && mobile && !android)
        iosShowSpecialOffer(placement);
        #end
    }

    // Is Offerwall Ready //

    private static function isOfferwallReady(placement:String = DEFAULT_PLACEMENT):Bool
    {
        if(!isInitialized) init();

        var ready:Bool = false;

        #if android
        if(androidIsOfferwallReady == null)
        {
            androidIsOfferwallReady = JNI.createStaticMethod(EXCLASS, "isOfferwallReady", "(Ljava/lang/String;)Z", true);
        }

        ready = androidIsOfferwallReady([placement]);
        #end

        #if(cpp && mobile && !android)
        ready = iosIsOfferwallReady(placement);
        #end

        return ready;
    }

    // Show Offerwall //

    private static function showOfferwall(placement:String = DEFAULT_PLACEMENT):Void
    {
        if(!isInitialized) init();

        #if android
        if(androidShowOfferwall == null)
        {
            androidShowOfferwall = JNI.createStaticMethod(EXCLASS, "showOfferwall", "(Ljava/lang/String;)V", true);
        }

        androidShowOfferwall([placement]);
        #end

        #if(cpp && mobile && !android)
        iosShowOfferwall(placement);
        #end
    }

    // Offerwall - internal callback //

    public function onCurrencyGranted(amount:Int):Void
    {
        if(offerwallCallback != null) offerwallCallback(amount);
    }

    // Request Local Notification Permission //

    public static function requestLocalNotificationPermission():Void
    {
        if(!isInitialized) init();

        #if android
        if(androidRequestLocalNotificationPermission == null)
        {
            androidRequestLocalNotificationPermission = JNI.createStaticMethod(EXCLASS, "requestLocalNotificationPermission", "()V", true);
        }

        androidRequestLocalNotificationPermission();
        #end

        #if(cpp && mobile && !android)
        iosRequestLocalNotificationPermission();
        #end
    }

    public static function enableLocalNotification(title:String, message:String, delay:Int):Void
    {
        if(!isInitialized) init();

        if(!hasLocalNotificationPermission) return;

        #if android
        if(androidEnableLocalNotification == null)
        {
            androidEnableLocalNotification = JNI.createStaticMethod(EXCLASS, "enableLocalNotification", "(Ljava/lang/String;Ljava/lang/String;I)V", true);
        }

        var args:Array<Dynamic> = new Array<Dynamic>();
        args.push(title);
        args.push(message);
        args.push(delay);
        androidEnableLocalNotification(args);
        #end

        #if(cpp && mobile && !android)
        iosEnableLocalNotification(title, message, delay);
        #end
    }

    public static function disableLocalNotification():Void
    {
        if(!isInitialized) init();

        if(!hasLocalNotificationPermission) return;

        #if android
        if(androidDisableLocalNotification == null)
        {
            androidDisableLocalNotification = JNI.createStaticMethod(EXCLASS, "disableLocalNotification", "()V", true);
        }

        androidDisableLocalNotification();
        #end

        #if(cpp && mobile && !android)
        iosDisableLocalNotification();
        #end
    }

    // Local Notification - internal callbacks //

    public function onPermissionGranted():Void
    {
        hasLocalNotificationPermission = true;

        if(permissionCallbacks[PERMISSION_GRANTED] != null) permissionCallbacks[PERMISSION_GRANTED]();
    }

    public function onPermissionRefused():Void
    {
        hasLocalNotificationPermission = false;

        if(permissionCallbacks[PERMISSION_REFUSED] != null) permissionCallbacks[PERMISSION_REFUSED]();
    }

    // Log Simple Event //

    public static function logSimpleEvent(eventType:String):Void
    {
        if(!isInitialized) init();

        #if android
        if(androidLogSimpleEvent == null)
        {
            androidLogSimpleEvent = JNI.createStaticMethod(EXCLASS, "logSimpleEvent", "(Ljava/lang/String;)V", true);
        }

        androidLogSimpleEvent([eventType]);
        #end

        #if(cpp && mobile && !android)
        iosLogSimpleEvent(eventType);
        #end
    }

    // Log Event //

    public static function logEvent(eventType:String, paramKey:String, paramValue:String):Void
    {
        if(!isInitialized) init();

        #if android
        if(androidLogEvent == null)
        {
            androidLogEvent = JNI.createStaticMethod(EXCLASS, "logEvent", "(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V", true);
        }

        androidLogEvent([eventType, paramKey, paramValue]);
        #end

        #if(cpp && mobile && !android)
        iosLogEvent(eventType, paramKey, paramValue);
        #end
    }

    // Is Purchasing Supported

    public static function isPurchasingSupported():Bool
    {
        if(!isInitialized) init();

        var supported:Bool = false;

        #if android
        if(androidIsPurchasingSupported == null)
        {
            androidIsPurchasingSupported = JNI.createStaticMethod(EXCLASS, "isPurchasingSupported", "()Z", true);
        }

        supported = androidIsPurchasingSupported();
        #end

        #if(cpp && mobile && !android)
        supported = iosIsPurchasingSupported();
        #end

        return supported;
    }

    // Attempt Purchase //

    private static function attemptPurchase(sku:String):Void
    {
        if(!isInitialized) init();

        #if android
        if(androidAttemptPurchase == null)
        {
            androidAttemptPurchase = JNI.createStaticMethod(EXCLASS, "attemptPurchase", "(Ljava/lang/String;)V", true);
        }

        androidAttemptPurchase([sku]);
        #end

        #if(cpp && mobile && !android)
        iosAttemptPurchase(sku);
        #end
    }

    // Purchase - internal callbacks //

    public function onPurchaseSuccess()
    {
        if(purchaseCallbacks[PURCHASE_SUCCESS] != null) purchaseCallbacks[PURCHASE_SUCCESS]();
    }

    public function onPurchaseFailed()
    {
        if(purchaseCallbacks[PURCHASE_FAILED] != null) purchaseCallbacks[PURCHASE_FAILED]();
    }

    // Get Display Price //

    public static function getDisplayPrice(sku:String, defaultPrice:String):String
    {
        if(!isInitialized) init();

        var displayPrice:String = defaultPrice;

        #if android
        if(androidGetDisplayPrice == null)
        {
            androidGetDisplayPrice = JNI.createStaticMethod(EXCLASS, "getDisplayPrice", "(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;", true);
        }

        displayPrice = androidGetDisplayPrice([sku, defaultPrice]);
        #end

        #if(cpp && mobile && !android)
        displayPrice = iosGetDisplayPrice(sku, displayPrice);
        #end

        return displayPrice;
    }

    // Is Item Owned //

    public static function isItemOwned(sku:String):Bool
    {
        if(!isInitialized) init();

        var owned:Bool = false;

        #if android
        if(androidIsItemOwned == null)
        {
            androidIsItemOwned = JNI.createStaticMethod(EXCLASS, "isItemOwned", "(Ljava/lang/String;)Z", true);
        }

        owned = androidIsItemOwned([sku]);
        #end

        #if(cpp && mobile && !android)
        owned = iosIsItemOwned(sku);
        #end

        return owned;
    }

    // Get Owned Item Count

    public static function getOwnedItemCount(sku:String):Int
    {
        if(!isInitialized) init();

        var count:Int = 0;

        #if android
        if(androidGetOwnedItemCount == null)
        {
            androidGetOwnedItemCount = JNI.createStaticMethod(EXCLASS, "getOwnedItemCount", "(Ljava/lang/String;)I", true);
        }

        count = androidGetOwnedItemCount([sku]);
        #end

        #if(cpp && mobile && !android)
        count = iosGetOwnedItemCount(sku);
        #end

        return count;
    }

    // Consume

    private static function consume(sku:String):Void
    {
        if(!isInitialized) init();

        #if android
        if(androidConsume == null)
        {
            androidConsume = JNI.createStaticMethod(EXCLASS, "consume", "(Ljava/lang/String;)V", true);
        }

        androidConsume([sku]);
        #end
        
        #if(cpp && mobile && !android)
        iosConsume(sku);
        #end
    }

    // Consume - internal callbacks

    public function onConsumeSuccess():Void
    {
        if(consumeCallbacks[CONSUME_SUCCESS] != null) consumeCallbacks[CONSUME_SUCCESS]();
    }

    public function onConsumeFailed():Void
    {
        if(consumeCallbacks[CONSUME_FAILED] != null) consumeCallbacks[CONSUME_FAILED]();
    }

    // Manually restore purchases

    public static function manuallyRestorePurchases():Void
    {
        if(!isInitialized) init();

        #if android
        if(androidManuallyRestorePurchases == null)
        {
            androidManuallyRestorePurchases = JNI.createStaticMethod(EXCLASS, "manuallyRestorePurchases", "()V", true);
        }

        androidManuallyRestorePurchases();
        #end

        #if(cpp && mobile && !android)
        iosManuallyRestorePurchases();
        #end
    }

    // Restore - internal callbacks

    public function onRestoreSuccess():Void
    {
        if(restoreCallbacks[RESTORE_SUCCESS] != null) restoreCallbacks[RESTORE_SUCCESS]();
    }

    public function onRestoreFailed():Void
    {
        if(restoreCallbacks[RESTORE_FAILED] != null) restoreCallbacks[RESTORE_FAILED]();
    }

    // Get display title

    public static function getDisplayTitle(sku:String, defaultTitle:String):String
    {
        if(!isInitialized) init();

        var result:String = defaultTitle;

        #if android
        if(androidGetDisplayTitle == null)
        {
            androidGetDisplayTitle = JNI.createStaticMethod(EXCLASS, "getDisplayTitle", "(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;", true);
        }

        result = androidGetDisplayTitle([sku, defaultTitle]);
        #end

        #if(cpp && mobile && !android)
        result = iosGetDisplayTitle(sku, defaultTitle);
        #end

        return result;
    }

    // Get display description

    public static function getDisplayDescription(sku:String, defaultDescription:String):String
    {
        if(!isInitialized) init();

        var result:String = defaultDescription;

        #if android
        if(androidGetDisplayDescription == null)
        {
            androidGetDisplayDescription = JNI.createStaticMethod(EXCLASS, "getDisplayDescription", "(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;", true);
        }

        result = androidGetDisplayDescription([sku, defaultDescription]);
        #end

        #if(cpp && mobile && !android)
        result = iosGetDisplayDescription(sku, defaultDescription);
        #end

        return result;
    }

    // GDPR

    public static function requiresDataConsentOptIn():Void
    {
        if(!isInitialized) init();

        #if android
        if(androidRequiresDataConsentOptIn == null)
        {
            androidRequiresDataConsentOptIn = JNI.createStaticMethod(EXCLASS, "requiresDataConsentOptIn", "()V");
        }
        androidRequiresDataConsentOptIn();
        #end

        #if(cpp && mobile && !android)
        iosRequiresDataConsentOptIn();
        #end
    }

    public static function serviceTermsOptIn(?requestedSdks:Array<Dynamic>):Void
    {
        if(!isInitialized) init();

        var requestedSdksAsString:String = null;
        if(requestedSdks != null && requestedSdks.length > 0)
        {
            var requestedSdksAsStringArray:Array<String> = cast requestedSdks;
            requestedSdksAsString = requestedSdksAsStringArray.join(",");
        }
        
        #if android
        if(androidServiceTermsOptIn == null)
        {
            androidServiceTermsOptIn = JNI.createStaticMethod(EXCLASS, "serviceTermsOptIn", "(Ljava/lang/String;)V", true);
        }
        androidServiceTermsOptIn([requestedSdksAsString]);
        #end

        #if(cpp && mobile && !android)
        iosServiceTermsOptIn(requestedSdksAsString);
        #end
    }

    public static function showServiceOptInDialogs(?requestedSdks:Array<Dynamic>):Void
    {
        if(!isInitialized) init();

        var requestedSdksAsString:String = null;
        if(requestedSdks != null && requestedSdks.length > 0)
        {
            var requestedSdksAsStringArray:Array<String> = cast requestedSdks;
            requestedSdksAsString = requestedSdksAsStringArray.join(",");
        }

        #if android
        if(androidShowServiceOptInDialogs == null)
        {
            androidShowServiceOptInDialogs = JNI.createStaticMethod(EXCLASS, "showServiceOptInDialogs", "(Ljava/lang/String;)V", true);
        }
        androidShowServiceOptInDialogs([requestedSdksAsString]);
        #end

        #if(cpp && mobile && !android)
        iosShowServiceOptInDialogs(requestedSdksAsString);
        #end
    }

    public static function serviceTermsOptOut():Void
    {
        if(!isInitialized) init();

        #if android
        if(androidServiceTermsOptOut == null)
        {
            androidServiceTermsOptOut = JNI.createStaticMethod(EXCLASS, "serviceTermsOptOut", "()V");
        }
        androidServiceTermsOptOut();
        #end

        #if(cpp && mobile && !android)
        iosServiceTermsOptOut();
        #end
    }

    public function onServiceOptInRequirement(required:Bool):Void
    {
        if(required && gdprCallbacks[ON_USER_OPT_IN_REQUIRED] != null) gdprCallbacks[ON_USER_OPT_IN_REQUIRED]();
        else if(!required && gdprCallbacks[ON_USER_OPT_IN_NOT_REQUIRED] != null) gdprCallbacks[ON_USER_OPT_IN_NOT_REQUIRED]();
    }

    public function onDialogComplete():Void
    {
        if(gdprCallbacks[ON_DIALOG_COMPLETE] != null) gdprCallbacks[ON_DIALOG_COMPLETE]();
    }

    #if android
    // Interstitial ad
    private static var androidIsInterstitialReady:Dynamic;
    private static var androidShowInterstitialAd:Dynamic;
    // Rewarded ad
    private static var androidIsRewardedAdReady:Dynamic;
    private static var androidShowRewardedAd:Dynamic;
    // Banner ad
    private static var androidIsBannerAdReady:Dynamic;
    private static var androidShowBannerAdWithPosition:Dynamic;
    private static var androidHideBannerAd:Dynamic;
    // Special Offer
    private static var androidIsSpecialOfferReady:Dynamic;
    private static var androidShowSpecialOffer:Dynamic;
    // Offerwall
    private static var androidIsOfferwallReady:Dynamic;
    private static var androidShowOfferwall:Dynamic;
    // Local Notifications
    private static var androidRequestLocalNotificationPermission:Dynamic;
    private static var androidEnableLocalNotification:Dynamic;
    private static var androidDisableLocalNotification:Dynamic;
    // Analytics
    private static var androidLogSimpleEvent:Dynamic;
    private static var androidLogEvent:Dynamic;
    // In-App Purchases
    private static var androidIsPurchasingSupported:Dynamic;
    private static var androidAttemptPurchase:Dynamic;
    private static var androidGetDisplayPrice:Dynamic;
    private static var androidIsItemOwned:Dynamic;
    private static var androidGetOwnedItemCount:Dynamic;
    private static var androidConsume:Dynamic;
    private static var androidManuallyRestorePurchases:Dynamic;
    private static var androidGetDisplayTitle:Dynamic;
    private static var androidGetDisplayDescription:Dynamic;
    // GDPR
    private static var androidRequiresDataConsentOptIn:Dynamic;
    private static var androidServiceTermsOptIn:Dynamic;
    private static var androidShowServiceOptInDialogs:Dynamic;
    private static var androidServiceTermsOptOut:Dynamic;
    #end

    #if(cpp && mobile && !android)
    // Interstitial Ad
    private static var iosIsInterstitialReady = CPPLoader.load(EXNAMESPACE, "ios_isInterstitialReady", 1);
    private static var iosShowInterstitialAd = CPPLoader.load(EXNAMESPACE, "ios_showInterstitialAd", 1);
    // Rewarded Ad
    private static var iosIsRewardedAdReady = CPPLoader.load(EXNAMESPACE, "ios_isRewardedAdReady", 1);
    private static var iosShowRewardedAd = CPPLoader.load(EXNAMESPACE, "ios_showRewardedAd", 1);
    // Banner Ad
    private static var iosIsBannerAdReady = CPPLoader.load(EXNAMESPACE, "ios_isBannerAdReady", 1);
    private static var iosShowBannerAdWithPosition = CPPLoader.load(EXNAMESPACE, "ios_showBannerAdWithPosition", 2);
    private static var iosHideBannerAd = CPPLoader.load(EXNAMESPACE, "ios_hideBannerAd", 0);
    // Special Offer
    private static var iosIsSpecialOfferReady = CPPLoader.load(EXNAMESPACE, "ios_isSpecialOfferReady", 1);
    private static var iosShowSpecialOffer = CPPLoader.load(EXNAMESPACE, "ios_showSpecialOffer", 1);
    // Offerwall
    private static var iosIsOfferwallReady = CPPLoader.load(EXNAMESPACE, "ios_isOfferwallReady", 1);
    private static var iosShowOfferwall = CPPLoader.load(EXNAMESPACE, "ios_showOfferwall", 1);
    // Local Notifications
    private static var iosRequestLocalNotificationPermission = CPPLoader.load(EXNAMESPACE, "ios_requestLocalNotificationPermission", 0);
    private static var iosEnableLocalNotification = CPPLoader.load(EXNAMESPACE, "ios_enableLocalNotification", 3);
    private static var iosDisableLocalNotification = CPPLoader.load(EXNAMESPACE, "ios_disableLocalNotification", 0);
    // Analytics
    private static var iosLogSimpleEvent = CPPLoader.load(EXNAMESPACE, "ios_logSimpleEvent", 1);
    private static var iosLogEvent = CPPLoader.load(EXNAMESPACE, "ios_logEvent", 3);
    // In-App Purchases
    private static var iosIsPurchasingSupported = CPPLoader.load(EXNAMESPACE, "ios_isPurchasingSupported", 0);
    private static var iosAttemptPurchase = CPPLoader.load(EXNAMESPACE, "ios_attemptPurchase", 1);
    private static var iosGetDisplayPrice = CPPLoader.load(EXNAMESPACE, "ios_getDisplayPrice", 2);
    private static var iosIsItemOwned = CPPLoader.load(EXNAMESPACE, "ios_isItemOwned", 1);
    private static var iosGetOwnedItemCount = CPPLoader.load(EXNAMESPACE, "ios_getOwnedItemCount", 1);
    private static var iosConsume = CPPLoader.load(EXNAMESPACE, "ios_consume", 1);
    private static var iosManuallyRestorePurchases = CPPLoader.load(EXNAMESPACE, "ios_manuallyRestorePurchases", 0);
    private static var iosGetDisplayTitle = CPPLoader.load(EXNAMESPACE, "ios_getDisplayTitle", 2);
    private static var iosGetDisplayDescription = CPPLoader.load(EXNAMESPACE, "ios_getDisplayDescription", 2);
    // GDPR
    private static var iosRequiresDataConsentOptIn = CPPLoader.load(EXNAMESPACE, "ios_requiresDataConsentOptIn", 0);
    private static var iosServiceTermsOptIn = CPPLoader.load(EXNAMESPACE, "ios_serviceTermsOptIn", 1);
    private static var iosShowServiceOptInDialogs = CPPLoader.load(EXNAMESPACE, "ios_showServiceOptInDialogs", 1);
    private static var iosServiceTermsOptOut = CPPLoader.load(EXNAMESPACE, "ios_serviceTermsOptOut", 0);
    #end
}
