/*
 * Enhance Stencyl Extension [Java backend]
 *
 * Created by Enhance, Inc.
 * www.enhance.co
 *
 */

package co.enhance;

import org.haxe.extension.Extension;
import org.haxe.lime.HaxeObject;

import android.util.Log;
import java.util.Arrays;
import java.util.List;

import co.enhance.Enhance;
import co.enhance.Enhance.*;
import co.enhance.EnhanceInAppPurchases;
import co.enhance.EnhanceInAppPurchases.*;

public class EnhanceExtension extends Extension 
{
    private static final String TAG = "EnhanceExtension";
    public static HaxeObject callbackObj = null;

    public EnhanceExtension() {}

    public static void init(HaxeObject haxeObject)
    {
        // this is the Haxe Enhance class instance (used for Java<->Haxe communication)
        callbackObj = haxeObject;

        // set offerwall callback

        CurrencyCallback currencyCallback = new CurrencyCallback()
        {
            @Override
            public void onCurrencyGranted(int currencyAmount)
            {
                final int currencyAmountInt = currencyAmount;

                Extension.callbackHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if(EnhanceExtension.callbackObj != null)
                            EnhanceExtension.callbackObj.call("onCurrencyGranted", new Object[] { currencyAmountInt });
                    }
                });
            }
        };
        Enhance.setReceivedCurrencyCallback(currencyCallback);
    }

    // Interstitial Ad //

    public static boolean isInterstitialReady(String placement)
    {
        if(placement == null) placement = Enhance.INTERSTITIAL_PLACEMENT_DEFAULT;
        return Enhance.isInterstitialReady(placement);
    }

    public static void showInterstitialAd(String placement)
    {
        if(placement == null) placement = Enhance.INTERSTITIAL_PLACEMENT_DEFAULT;
        Enhance.showInterstitialAd(placement);
    }

    // Rewarded Ad //

    public static boolean isRewardedAdReady(String placement)
    {
        if(placement == null) placement = Enhance.REWARDED_PLACEMENT_NEUTRAL;
        return Enhance.isRewardedAdReady(placement);
    }

    public static void showRewardedAd(String placement) 
    {
        final RewardCallback rewardCallbacks = new RewardCallback() 
        {
            @Override
            public void onRewardGranted(int rewardValue, RewardType rewardType) 
            {
                // The values which we pass to Haxe need to be "final", it will not compile otherwise.
                final int rewardValueInt = rewardValue; 
                final String rewardTypeStr;
                if(rewardType == RewardType.COINS) rewardTypeStr = "COINS";
                else rewardTypeStr = "ITEM";

                Extension.callbackHandler.post(new Runnable() 
                {
                    @Override 
                    public void run() 
                    {
                        if(EnhanceExtension.callbackObj != null)
                            EnhanceExtension.callbackObj.call("onRewardGranted", new Object[] { rewardTypeStr, rewardValueInt });
                    }
                });
            }

            @Override
            public void onRewardDeclined()
            {
                Extension.callbackHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if(EnhanceExtension.callbackObj != null)
                            EnhanceExtension.callbackObj.call("onRewardDeclined", new Object[] {});
                    }
                });
            }

            @Override
            public void onRewardUnavailable()
            {
                Extension.callbackHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if(EnhanceExtension.callbackObj != null)
                            EnhanceExtension.callbackObj.call("onRewardUnavailable", new Object[] {});
                    }
                });
            }
        };

        if(placement == null) placement = Enhance.REWARDED_PLACEMENT_NEUTRAL;
        Enhance.showRewardedAd(rewardCallbacks, placement);
    } 

    // Banner Ad //

    public static boolean isBannerAdReady(String placement)
    {
        if(placement == null) placement = Enhance.PLACEMENT_DEFAULT;
        return Enhance.isBannerAdReady(placement);
    }

    public static void showBannerAdWithPosition(String placement, String positionStr)
    {
        if(positionStr == null) positionStr = "BOTTOM";

        Position position = Position.BOTTOM;
        if(positionStr.equals("TOP")) position = Position.TOP;

        if(placement == null) placement = Enhance.PLACEMENT_DEFAULT;
        if(position == null) position = Position.BOTTOM;

        Enhance.showBannerAdWithPosition(placement, position);
    }

    public static void hideBannerAd()
    {
        Enhance.hideBannerAd();
    }

    // Special Offer //

    public static boolean isSpecialOfferReady(String placement)
    {
        if(placement == null) placement = Enhance.PLACEMENT_DEFAULT;
        return Enhance.isSpecialOfferReady(placement);
    }

    public static void showSpecialOffer(String placement)
    {
        if(placement == null) placement = Enhance.PLACEMENT_DEFAULT;
        Enhance.showSpecialOffer(placement);
    }

    // Offerwall //

    public static boolean isOfferwallReady(String placement)
    {
        if(placement == null) placement = Enhance.PLACEMENT_DEFAULT;
        return Enhance.isOfferwallReady(placement);
    }

    public static void showOfferwall(String placement)
    {
        if(placement == null) placement = Enhance.PLACEMENT_DEFAULT;
        Enhance.showOfferwall(placement);
    }

    // Local Notifications //

    public static void requestLocalNotificationPermission()
    {
        PermissionCallback permissionCallbacks = new PermissionCallback()
        {
            @Override
            public void onPermissionGranted()
            {
                Extension.callbackHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if(EnhanceExtension.callbackObj != null)
                            EnhanceExtension.callbackObj.call("onPermissionGranted", new Object[] {});
                    }
                });
            }

            @Override
            public void onPermissionRefused()
            {
                Extension.callbackHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if(EnhanceExtension.callbackObj != null)
                            EnhanceExtension.callbackObj.call("onPermissionRefused", new Object[] {});
                    }
                });
            }
        };

        Enhance.requestLocalNotificationPermission(permissionCallbacks);
    }

    public static void enableLocalNotification(String title, String message, int delay)
    {
        if(title == null || message == null) return; 
        Enhance.enableLocalNotification(title, message, delay);
    }

    public static void disableLocalNotification() 
    {
        Enhance.disableLocalNotification();
    }

    // Analytics //

    public static void logSimpleEvent(String eventType)
    {
        if(eventType == null) return;
        Enhance.logEvent(eventType);
    }

    public static void logEvent(String eventType, String paramKey, String paramValue)
    {
        if(eventType == null || paramKey == null || paramValue == null) return;
        Enhance.logEvent(eventType, paramKey, paramValue);
    }

    // Is Purchasing Supported //

    public static boolean isPurchasingSupported()
    {
        return Enhance.purchases.isSupported();
    }

    // Attempt Purchase //

    public static void attemptPurchase(String sku)
    {
        if(sku == null) return;

        PurchaseCallback purchaseCalbacks = new PurchaseCallback()
        {
            @Override 
            public void onPurchaseSuccess()
            {
                Extension.callbackHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if(EnhanceExtension.callbackObj != null)
                            EnhanceExtension.callbackObj.call("onPurchaseSuccess", new Object[] {});
                    }
                });
            }

            @Override
            public void onPurchaseFailed()
            {
                Extension.callbackHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if(EnhanceExtension.callbackObj != null)
                            EnhanceExtension.callbackObj.call("onPurchaseFailed", new Object[] {});
                    }
                });
            }
        };
        Enhance.purchases.attemptPurchase(sku, purchaseCalbacks);
    }

    // Get Display Price //

    public static String getDisplayPrice(String sku, String defaultPrice)
    {
        if(sku == null || defaultPrice == null) return "";

        return Enhance.purchases.getDisplayPrice(sku, defaultPrice);
    }

    // Is Item Owned //

    public static boolean isItemOwned(String sku)
    {
        if(sku == null) return false;

        return Enhance.purchases.isItemOwned(sku);
    }

    // Get Owned Item Count //

    public static int getOwnedItemCount(String sku)
    {
        if(sku == null) return 0;

        return Enhance.purchases.getOwnedItemCount(sku);
    }

    // Consume //

    public static void consume(String sku)
    {
        if(sku == null) return;

        ConsumeCallback consumeCallbacks = new ConsumeCallback()
        {
            @Override
            public void onConsumeSuccess()
            {
                Extension.callbackHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if(EnhanceExtension.callbackObj != null)
                            EnhanceExtension.callbackObj.call("onConsumeSuccess", new Object[] {});
                    }
                });
            }

            @Override
            public void onConsumeFailed()
            {
                Extension.callbackHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if(EnhanceExtension.callbackObj != null)
                            EnhanceExtension.callbackObj.call("onConsumeFailed", new Object[] {});
                    }
                });
            }
        };
        Enhance.purchases.consume(sku, consumeCallbacks);
    }

    // Manually restore purchases //

    public static void manuallyRestorePurchases()
    {
        RestoreCallback restoreCallbacks = new RestoreCallback()
        {
            @Override
            public void onRestoreSuccess()
            {
                Extension.callbackHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if(EnhanceExtension.callbackObj != null)
                            EnhanceExtension.callbackObj.call("onRestoreSuccess", new Object[] {});
                    }
                });
            }

            @Override
            public void onRestoreFailed()
            {
                Extension.callbackHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if(EnhanceExtension.callbackObj != null)
                            EnhanceExtension.callbackObj.call("onRestoreFailed", new Object[] {});
                    }
                });
            }
        };
        Enhance.purchases.manuallyRestorePurchases(restoreCallbacks);
    }

    // Get Display Title //

    public static String getDisplayTitle(String sku, String defaultTitle)
    {
        if(sku == null || defaultTitle == null) return "";

        return Enhance.purchases.getDisplayTitle(sku, defaultTitle);
    }

    // Get Display Description //

    public static String getDisplayDescription(String sku, String defaultDescription)
    {
        if(sku == null || defaultDescription == null) return "";

        return Enhance.purchases.getDisplayDescription(sku, defaultDescription);
    }

    // GDPR

    public static void requiresDataConsentOptIn() {
        OptInRequiredCallback callback = new OptInRequiredCallback() {
            @Override
            public void onServiceOptInRequirement(final boolean isUserOptInRequired) {
                Extension.callbackHandler.post(new Runnable() {
                    @Override
                    public void run()
                    {
                        if(EnhanceExtension.callbackObj != null)
                            EnhanceExtension.callbackObj.call("onServiceOptInRequirement", new Object[] { isUserOptInRequired });
                    }
                });
            }
        };
        Enhance.requiresDataConsentOptIn(callback);
    }

    public static void serviceTermsOptIn(String requestedSdksString) {
        if(requestedSdksString != null && !requestedSdksString.isEmpty()) {
            String[] array = requestedSdksString.split(",");
            List<String> requestedSdks = Arrays.asList(array);
            Enhance.serviceTermsOptIn(requestedSdks);
            return;
        }

        Enhance.serviceTermsOptIn();
    }

    public static void serviceTermsOptOut() {
        Enhance.serviceTermsOptOut();
    }

    public static void showServiceOptInDialogs(String requestedSdksString) {
         OnDataConsentOptInComplete callback = new OnDataConsentOptInComplete() {
            @Override
            public void onDialogComplete() {
                Extension.callbackHandler.post(new Runnable() {
                    @Override
                    public void run()
                    {
                        if(EnhanceExtension.callbackObj != null)
                            EnhanceExtension.callbackObj.call("onDialogComplete", new Object[] {});
                    }
                });
            }
        };
        Enhance.showServiceOptInDialogs(callback);
    }
}
